/****************************************************************************************************\
 *                                                                                                  *
\****************************************************************************************************/

#include <GL/glew.h>
// include some convenience functions (names begin with "ue_" to distinguish them from CUDA)
#include "utility_environment.h"

// OpenGL utilities
#include "opengl_includes.h"

#include <thrust/swap.h>

/// constant configuration
#define SPREAD 0.2
#define SQRT_SPREAD_RAYS 16
#define MAX_REFLECTIONS 10

////////////////////////////////////////////////////////////////////////////////
// GLUT, which we use for displaying images, makes use of callback functions. //
// The following global variables are the simplest way of passing parameters  //
// between the various callback functions.                                    //
////////////////////////////////////////////////////////////////////////////////

// keep track of mouse state
int mouseButtonLeft = GLUT_UP;
int mouseButtonMiddle = GLUT_UP;
int mouseButtonRight = GLUT_UP;
int mouseLastX = 0;
int mouseLastY = 0;

const int imgWidth = 2*768;
const int imgHeight = 2*512;

CPUBitmap *g_bitmap = 0;     // image for display

struct Camera;
Camera *g_cam = 0;           // virtual camera

pixelRGBA8888 *Image;         // buffer for floating point pixels, in host memory
pixelRGBA8888 *d_Image;       // buffer for floating point pixels, in device memory

struct Sphere;
const int g_numSpheres = 3;    // number of spheres
Sphere *g_spheres;             // storage space for the spheres, host


/***********************\
 *  vector arithmetic  *
\***********************/
struct Vec3
{
	float x;
	float y;
	float z;

	__device__ __host__
	Vec3()      // empty default constructor
	{
	}

	__device__ __host__
	Vec3(const float X, const float Y, const float Z)   // with initialization
	{
		x = X;
		y = Y;
		z = Z;
	}

	// arithmetic
	__device__ __host__
	Vec3 operator+(const Vec3 v) const     // vector add
	{
		return Vec3(x + v.x, y + v.y, z + v.z);
	}
  
	__device__ __host__
	Vec3 operator-(const Vec3 v) const     // vector subtract
	{
		return Vec3(x - v.x, y - v.y, z - v.z);
	}
  
	__device__ __host__
	Vec3 operator*(const float s) const    // scalar multiply
	{
		return Vec3(x*s, y*s, z*s);
	}
  
	__device__ __host__
	float dot(const Vec3 v) const          // dot product
	{
		return x*v.x + y*v.y + z*v.z;
	}

	__device__ __host__
	Vec3 cross(const Vec3 v) const         // cross product
	{
		return Vec3(y*v.z - z*v.y,
			z*v.x - x*v.z,
			x*v.y - y*v.x);
	}

	__device__ __host__
	float length(void) const               // vector magnitude
	{
		return sqrtf(dot(*this));
	}

	__device__ __host__
	void normalize(void) 
	{
		float len = length();
		x /= len;
		y /= len;
		z /= len;
	}
};

/***************\
 *  view rays  *
\***************/

struct Ray
{
	Vec3 base;
	Vec3 dir;
  
	__device__ __host__
	Ray(void)                                 // default constructor
	{
		base = Vec3(0.0f, 0.0f, 0.0f);
		dir = Vec3(0.0f, 0.0f, 0.0f);
	}

	__device__ __host__
	Ray(const Vec3& Base, const Vec3& Dir)    // initialization
	{
		base = Base;
		dir = Dir;
	}
};

/************\
 *  camera  *
\************/

// basic perspective camera
struct Camera
{
	Vec3 eye;          // eye point
	Vec3 sky;          // sky vector
	Vec3 viewCenter;   // center of view rect
	Vec3 viewRight;    // horizontal direction of view rect
	Vec3 viewUp;       // vertical direction of view rect
	int imgWidth;      // image width in pixels
	int imgHeight;     // image height in pixels

	Camera(void) {     // default constructor
		eye = Vec3(0.0f, 0.0f, 3.0f);        // 3 units in front of screen
		sky = Vec3(0.0f, 1.0f, 0.0f);        // sky, used to align viewUp
		viewCenter = Vec3(0.0f, 0.0f, 0.0f); // screen center at origin
		viewRight = Vec3(1.6f, 0.0f, 0.0f);  // to right edge of 16:10 screen
		viewUp = Vec3(0.0f, 1.0f, 0.0f);     // to top edge of 16:10 screen
	}

	// translate camera position
	// does not maintain point of interest, may need to lookAt() afterwards
	void move(Vec3 translation) {
		eye = eye + translation;
		viewCenter = viewCenter + translation;
	}

	void lookAt(Vec3 poi) {   // aim camera at point of interest
		// first derive some invariant scalar measures of the current camera
		Vec3 forward = viewCenter - eye;
		float viewWidth = viewRight.length();
		float viewHeight = viewUp.length();
		float distance = forward.length();
    
		// now rebuild the camera
		forward = poi - eye;
		forward.normalize();
		forward = forward * distance;
		viewCenter = eye + forward;        // from eye towards point of interest

		viewRight = forward.cross(sky);    // horizon is perpendicular to sky
		viewRight.normalize();
		viewRight = viewRight * viewWidth;

		viewUp = viewRight.cross(forward); // perpendicular to horizon and forward
		viewUp.normalize();
		viewUp = viewUp * viewHeight;
	}

	__device__ __host__
	Ray makePrimaryRay(int x, int y) {   // returns ray through center of pixel
		// start from relative pixel coordinates in [0.0 .. 1.0]
		// with y axis pointing down, pixel center is at (+0.5, +0.5)
		float fx = (float(x) + 0.5f) / float(imgWidth);
		float fy = (float(y) + 0.5f) / float(imgHeight);

		// map to [-1.0 .. 1.0], with y axis pointing down
		fx = 2.0f * fx - 1.0f;
		fy = 2.0f * fy - 1.0f;

		// ray direction is from eye to that point in view rect
		Vec3 dir = viewCenter + (viewRight * fx) + (viewUp * fy) - eye;
		dir.normalize();
		return Ray(eye, dir);
	}
};

/************\
 *  sphere  *
\************/

#define HUGENUM 1e20f      // some huge number limits traced distances

struct Sphere
{
	// geometry of sphere
	Vec3 center;
	float radius;

	// material
	Vec3 color;              // pigment
	float ambient;           // ambient amount
	float diffuse;           // diffuse amount
	float highlight;         // highlight intensity
	float roughness;         // controls highlight size (shinyness)
	float reflection;        // weight of reflected ray

	/*
	Sphere() {}
	~Sphere() {}
	*/

	// return smallest positive t
	__device__ __host__
	float intersect(const Ray& ray)
	{
		// translate sphere center to origin, i.e. translate ray base
		Vec3 base = ray.base - center;
    
		// quadratic coefficients
		// immediately divide out leading coefficient
		// and immediately multiply b by -0.5 to simplify further computation
		float a = ray.dir.dot(ray.dir);
		float b = base.dot(ray.dir)/-a;
		float c = (base.dot(base) - radius*radius)/a;

		// so now we have to solve:   t*t - 2.0*t*b + c = 0;
		a = b*b - c;  // discriminant
    
		float t = HUGENUM;    // initialize as "ray misses sphere"
    
		if (a >= 0.0f)
		{
			// if ray hits sphere
			a = sqrtf(a);
			float tmp = b - a;  // check near hit
			if (tmp > 0.0f) {
				// if near hit is > 0, keep it
				t = tmp;
			}
			else
			{
				tmp = b + a;      // check far hit
				if (tmp > 0.0f) {
					// if far hit is > 0, keep it
					t = tmp;
				}   
			}
		}
		return t;    // return hit or miss
	}

	// return normal vector at point, unit length
	__device__ __host__
	Vec3 normal(Vec3 where)
		{
		Vec3 n = where - center;
		n.normalize();
		return n;
	}
};

/****************\
 *  trace rays  *
\****************/

__constant__ Sphere c_spheres[g_numSpheres];             // storage for the spheres, in device's constant memory

// returns true if light ray is blocked
__device__
bool shadow(const Ray& ray, int numSpheres)
{
	for (int i = numSpheres - 1; i >= 0; i--)
	{
		if (c_spheres[i].intersect(ray) < HUGENUM)
		{
			return true;
		}
	}
  
	return false;
}


// returns color for ray
__device__
Vec3 trace(Ray& ray, int numSpheres, bool subtraces = true)
{
	Vec3 lightPos = Vec3(-1.0f, 2.0f, 1.0f);  // direction towards infinitely distant light source

	float weight = 1.0f; // weight of remaining reflected ray, if any
	int rayDepth = 0;    // number of reflections so far
	Vec3 result = Vec3(0.0f, 0.0f, 0.0f);  // accumulates ray intensities

	float floorDist;     // distance to floor
	int sphereIdx;       // closest valid sphere hit, if any
	float sphereDist;    // distance to closest valid sphere hit, if any

	Vec3 normal;         // normal vector at the current valid hit
	Vec3 position;       // world coordinates of the current valid hit
	Vec3 reflected;      // reflected direction at the current valid hit
	Vec3 color;          // color of the current ray (segment)

	float ambient;
	float diffuse;
	float highlight;
	float roughness;
	float reflection;

	while (rayDepth < MAX_REFLECTIONS)
	{
		rayDepth++;

		//skip tracing if the weight is too small 
		//-> further reflections not relevant 
		//-> color will not be changed in further iterations
		if (weight > 0.1f)
		{
			// test ground floor (at y = 0) first
			floorDist = HUGENUM;
			if (ray.dir.y != 0.0f)
			{
				floorDist = -ray.base.y / ray.dir.y;  // in terms of ray parameter t
				if (floorDist < 0.0f)
				{
					floorDist = HUGENUM;
				}
			}

			sphereIdx = -1;  // remember closest valid sphere hit
			sphereDist = HUGENUM;

			for (int i = numSpheres - 1; i >= 0; i--)  // test all spheres
			{
				float dist = c_spheres[i].intersect(ray);
				if (dist < sphereDist)
				{
					sphereDist = dist;             // remember closest hit
					sphereIdx = i;
				}
			}

			// initialize as default material (for floor)
			ambient = 0.3f;
			diffuse = 0.7f;
			highlight = 0.0f;
			roughness = 0.05f;
			reflection = 0.0f;

			// did we hit anything?  
			if (fminf(sphereDist, floorDist) >= HUGENUM)
			{
				// we missed the floor and all spheres, so it's sky dome color
				color = (Vec3(0.7f, 0.8f, 0.9f) * (1.0f - ray.dir.y)) +
					(Vec3(0.1f, 0.1f, 0.9f) * ray.dir.y);
				reflection = 0.0f;  // no further reflections off of sky dome
			}
			else
			{
				// what is closer, floor or sphere?
				if (floorDist < sphereDist)
				{
					// ray did hit the floor, make checkerboard pattern
					position = ray.base + (ray.dir * floorDist);
					float toggle = float((int(floorf(position.x))  // toggle distinguishes between "even" and "odd" squares
						^ int(floorf(position.z))) & 1);
					color = (Vec3(0.9f, 0.9f, 0.1f) * toggle) +
						(Vec3(0.1f, 0.9f, 0.1f) * (1.0f - toggle));
					normal = Vec3(0.0f, 1.0f, 0.0f);
				}
				else
				{
					// ray did hit a sphere
					color = c_spheres[sphereIdx].color;
					position = ray.base + (ray.dir * sphereDist);
					normal = c_spheres[sphereIdx].normal(position);

					// sphere material
					ambient = c_spheres[sphereIdx].ambient;
					diffuse = c_spheres[sphereIdx].diffuse;
					highlight = c_spheres[sphereIdx].highlight;
					roughness = c_spheres[sphereIdx].roughness;
					reflection = c_spheres[sphereIdx].reflection;
				}
				// light that point
				Vec3 lightDir = lightPos;
				lightDir.normalize();

				// reflected ray direction
				reflected = ray.dir - (normal * (2.0f*normal.dot(ray.dir)));
				reflected.normalize();

				// cheat to avoid erroneous self intersections
				position = position + (normal * 0.0005f);

				float brightness = ambient;  // start with amount of ambient light
				float cosAngle = 0.0f;    // temporary for diffuse and highlight
				if (!shadow(Ray(position, lightDir), numSpheres))
				{
					// if light source is visible, add its diffuse contribution
					cosAngle = lightDir.dot(normal);
					if (cosAngle > 0.0f)
					{
						brightness += diffuse*cosAngle;
					}

					// add specular higlight, i.e. a reflection of the light source	
					cosAngle = reflected.dot(lightDir);
					if (cosAngle > 0.0f)   // only if reflected direction is towards light
					{
						// resize highlight (shrinks with increasing factor)
						cosAngle = ((cosAngle - 1.0f) / roughness) + 1.0f;

						if (cosAngle > 0.0f)   // only if we are inside the highlight
						{
							cosAngle *= cosAngle;
							cosAngle *= highlight;  // highlight intensity
						}
						else
						{
							cosAngle = 0.0f;
						}
					}
					else
					{
						cosAngle = 0.0f;
					}
				}
				color = (color * brightness) + Vec3(cosAngle, cosAngle, cosAngle);
			}

			// Without scatter, consider simple reflected color
			// TODO: Only add direct material color during first iteration and for scattered rays
			result = result + (color * weight);
			weight *= reflection;    // diminish weight of further reflections

			// follow the reflected ray in next iteration
			ray.base = position;
			ray.dir = reflected;
		}

		if (subtraces) {
			// TODO: Calculate average color of all scattered rays using DP
			Vec3 scatteredAvgColor(0,0,0);
			result = result + scatteredAvgColor * weight;
		} else {
			return result;
		}
	}
	return result;
}

__global__
void render_gpu(Camera cam, pixelRGBA8888 *Image, int numSpheres) {
	int x = blockDim.x * blockIdx.x + threadIdx.x;
	int y = blockDim.y * blockIdx.y + threadIdx.y;

	int imgWidth = cam.imgWidth;
	int imgHeight = cam.imgHeight;

	if (y < imgHeight)
	{
		if (x < imgWidth)
		{
			Ray ray = cam.makePrimaryRay(x,y);
			Vec3 color = trace(ray, numSpheres);

			Image[y * imgWidth + x].r = (unsigned char)(fminf(color.x, 1.0f)*255.f);
			Image[y * imgWidth + x].g = (unsigned char)(fminf(color.y, 1.0f)*255.f);
			Image[y * imgWidth + x].b = (unsigned char)(fminf(color.z, 1.0f)*255.f);
		}
	}
}

/******************************************\
 *  functions used during GLUT callbacks  *
\******************************************/

// compute and draw image
void computeImage(int imgWidth, int imgHeight, pixelRGBA8888 *Image)
{
	CudaTimer timer;
	float elapsedTime = 0.0f;
	
	// set threads layout;  we utilize the fact that img_size = k*32 (see above)
	// consequence: blocks.x * threads.x = img_size
	// -> we don't need to check in the kernel whether the thread is outside the image!
	// image is subdivided into rectangular tiles for parallelism - this variable controls tile size
    dim3 cudaBlockSize(16,16);  
    dim3 cudaGridSize((imgWidth + cudaBlockSize.x - 1)/cudaBlockSize.x, (imgHeight + cudaBlockSize.y - 1)/cudaBlockSize.y);
	// TODO: Allocate shared memory based on your need
	size_t sharedMemorySize = 0;

	timer.start();
	render_gpu <<< cudaGridSize, cudaBlockSize, sharedMemorySize >>> (*g_cam, d_Image, g_numSpheres);
	cudaDeviceSynchronize();  // wait until GPU is finished
	handleCUDAerror(__LINE__);

	// copy GPU result back to host memory
    cudaMemcpy(Image, d_Image, imgWidth*imgHeight*sizeof(pixelRGBA8888), cudaMemcpyDeviceToHost);
	handleCUDAerror(__LINE__);

	elapsedTime = timer.elapsed();
	printf("calculation finished after %5.2f milliseconds.\n", elapsedTime);
}

void cleanUp()
{
	delete g_bitmap;
	delete[] Image;
}

// static methods used for glut callbacks

void mouseButton(int button, int state, int x, int y)
{
	mouseLastX = x;  // remember mouse position
	mouseLastY = y;
	
	switch(button) { // remember mouse button state
	case GLUT_LEFT_BUTTON:
		mouseButtonLeft = state;
		break;
	case GLUT_MIDDLE_BUTTON:
		mouseButtonMiddle = state;
		break;

	case GLUT_RIGHT_BUTTON:
		mouseButtonRight = state;
		break;
  }
}

void mouseDrag(int x, int y)
{
	const float aimSpeed = 0.005f;   // mouse sensitivity while aiming
	const float moveSpeed = 0.01f;   // mouse sensitivity while moving
  
	int deltaX = x - mouseLastX;  // relative mouse movement
	int deltaY = y - mouseLastY;
  
	mouseLastX = x;  // remember mouse position
	mouseLastY = y;
  
	Vec3 dirY = g_cam->viewUp;    // camera relative directions in world space
	dirY.normalize();
	Vec3 dirX = g_cam->viewRight;
	dirX.normalize();
  
	if (mouseButtonLeft == GLUT_DOWN)
	{
		Vec3 target = g_cam->viewCenter;
		float fx = float(deltaX) * aimSpeed;
		float fy = -float(deltaY) * aimSpeed;
		target = target + (dirY * fy) + (dirX * fx);
		g_cam->lookAt(target);
	}
  
	if (mouseButtonRight == GLUT_DOWN)
	{
		float fx = float(deltaX) * moveSpeed;
		float fy = float(-deltaY) * moveSpeed;
		Vec3 offset = (dirY * fy) + (dirX * fx);
		g_cam->move(offset);
	}

	if (mouseButtonMiddle == GLUT_DOWN)
	{
		Vec3 forward = g_cam->viewCenter - g_cam->eye;
		forward.normalize();
		g_cam->move(forward * (0.05f * float(-deltaY)));
	}

	glutPostRedisplay();
}

void keyboard(unsigned char key, int x, int y)
{
	switch (key) 
	{
	case 27:              // keycode for the Escape key
		cleanUp();
		exit(0);
		break;
	}
}

void specialKeyboard(int key, int x, int y)
{
	const float turnSpeed = 0.075f;
	const float moveSpeed = 0.1f;

	Vec3 center = g_cam->viewCenter;
	Vec3 forward = center - g_cam->eye;
	float distance = forward.length();
	forward.normalize();
	Vec3 direction;

	switch (key) 
	{
	case GLUT_KEY_LEFT:              // cursor left key:
		direction = g_cam->viewRight;
		direction.normalize();
		direction = direction * (distance * turnSpeed);
		g_cam->lookAt(center - direction);
		break;
	case GLUT_KEY_RIGHT:             // cursor right key:
		direction = g_cam->viewRight;
		direction.normalize();
		direction = direction * (distance * turnSpeed);
		g_cam->lookAt(center + direction);
		break;
	case GLUT_KEY_UP:                // cursor up key:
		direction = g_cam->viewUp;
		direction.normalize();
		direction = direction * (distance * turnSpeed);
		g_cam->lookAt(center + direction);
		break;
	case GLUT_KEY_DOWN:              // cursor down key:
		direction = g_cam->viewUp;
		direction.normalize();
		direction = direction * (distance * turnSpeed);
		g_cam->lookAt(center - direction);
		break;
	case GLUT_KEY_PAGE_UP:           // page up key:
		g_cam->move(forward * moveSpeed);
		break;
	case GLUT_KEY_PAGE_DOWN:         // page down key:
		g_cam->move(forward * (-moveSpeed));
		break;
	}
	glutPostRedisplay();
}

// calls computation, pixel conversion, and OpenGL draw when GLUT refreshes the display
void display( void ) 
{
	computeImage(imgWidth, imgHeight, g_bitmap->get_ptr());
	g_bitmap->draw();
	glFlush();
}


/************************
 *  Setup and Test Rig  *
\************************/
int main(int argc, char **argv)
{
	///////////////////////////////////////////////////////////////////////////
	// initialize GLUT, open display window, and register callback functions //
	///////////////////////////////////////////////////////////////////////////
	glutInit( &argc, argv );
	glutInitDisplayMode( GLUT_SINGLE | GLUT_RGBA );
	glutInitWindowSize( imgWidth, imgHeight );
	glutCreateWindow( "press ESC to close" );
	glutKeyboardFunc(keyboard);
	glutSpecialFunc(specialKeyboard);
	glutDisplayFunc(display);
	glutMouseFunc(mouseButton);
	glutMotionFunc(mouseDrag);

	// initialize necessary OpenGL extensions
	glewInit();
	if (! glewIsSupported( "GL_VERSION_2_0 " 
		"GL_ARB_pixel_buffer_object"
		)) 
	{
			fprintf( stderr, "ERROR: Support for necessary OpenGL extensions missing.");
			fflush( stderr);
			exit(1);
	}

	///////////
	// setup //
	///////////
	ue_checkGPU();                 // verify that a CUDA device is available
	
	// allocate buffers
	g_bitmap = new CPUBitmap(imgWidth, imgHeight);
	Image = new pixelRGBA8888[imgWidth * imgHeight];
	cudaMalloc(&d_Image, imgWidth*imgHeight*sizeof(pixelRGBA8888));
	g_spheres = new Sphere[g_numSpheres];
	g_cam = new Camera;

	// initialize scene
	g_spheres[0].center = Vec3(0.0f, 0.6f, -1.0f);
	g_spheres[0].radius = 0.6f;
	g_spheres[0].color = Vec3(0.9f, 0.1f, 0.1f);
	g_spheres[0].ambient = 0.1f;
	g_spheres[0].diffuse = 0.3f;
	g_spheres[0].highlight = 0.8f;
	g_spheres[0].roughness = 0.05f;
	g_spheres[0].reflection = 0.95f;

	g_spheres[1].center = Vec3(-1.0f, 0.5f, -2.0f);
	g_spheres[1].radius = 0.5f;
	g_spheres[1].color = Vec3(0.7f, 0.6f, 0.5f);
	g_spheres[1].ambient = 0.3f;
	g_spheres[1].diffuse = 0.7f;
	g_spheres[1].highlight = 0.8f;
	g_spheres[1].roughness = 0.05f;
	g_spheres[1].reflection = 0.0f;

	g_spheres[2].center = Vec3(1.0f, 0.4f, -1.0f);
	g_spheres[2].radius = 0.4f;
	g_spheres[2].color = Vec3(0.1f, 0.1f, 0.9f);
	g_spheres[2].ambient = 0.1f;
	g_spheres[2].diffuse = 0.3f;
	g_spheres[2].highlight = 0.8f;
	g_spheres[2].roughness = 0.05f;
	g_spheres[2].reflection = 0.95f;

	// copy spheres to device
	cudaMemcpyToSymbol( c_spheres, g_spheres, g_numSpheres * sizeof(Sphere) );

	// initialize camera
	g_cam->move(Vec3(-1.0f, 1.25f, 1.0f));
	g_cam->lookAt(Vec3(0.0f, 1.0f, 0.0f));
	g_cam->imgWidth = imgWidth;
	g_cam->imgHeight = imgHeight;

	// start rendering, see display()
	glutMainLoop();

	exit(EXIT_SUCCESS);
}
